@@ -2,9 +2,11 @@ import time
 import logging
 
 from django.core.cache import cache
+from django.conf import settings
 
 from cacheback import tasks
 
+logging.basicConfig()
 logger = logging.getLogger('cacheback')
 
 MEMCACHE_MAX_EXPIRATION = 2592000
@@ -23,23 +25,29 @@ class Job(object):
     # for refreshing this item.
 
     #: Default cache lifetime is 5 minutes.  After this time, the result will
-    #  be considered stale and requests will trigger a job to refresh it.
+    #: be considered stale and requests will trigger a job to refresh it.
     lifetime = 600
 
     #: Timeout period during which no new Celery tasks will be created for a
-    #  single cache item.  This time should cover the normal time required to
-    #  refresh the cache.
+    #: single cache item.  This time should cover the normal time required to
+    #: refresh the cache.
     refresh_timeout = 60
 
     #: Time to store items in the cache.  After this time, we will get a cache
-    #  miss which can lead to synchronous refreshes if you have
-    #  fetch_on_miss=True.
+    #: miss which can lead to synchronous refreshes if you have
+    #: fetch_on_miss=True.
     cache_ttl = MEMCACHE_MAX_EXPIRATION
 
-    # Default behaviour is to do a synchronous fetch when the cache is empty.
-    # Stale results are generally ok, but not no results.
+    #: Whether to perform a synchronous refresh when a result is missing from
+    #: the cache.  Default behaviour is to do a synchronous fetch when the cache is empty.
+    #: Stale results are generally ok, but not no results.
     fetch_on_miss = True
 
+    #: Whether to perform a synchronous refresh when a result is in the cache
+    #: but stale from. Default behaviour is never to do a synchronous fetch but
+    #: there will be times when an item is _too_ stale to be returned.
+    fetch_on_stale_threshold = None
+
     # --------
     # MAIN API
     # --------
@@ -51,10 +59,11 @@ class Job(object):
         This method is not intended to be overidden
         """
         # We pass args and kwargs through a filter to allow them to be
-        # converted into values that can be picked.
+        # converted into values that can be pickled.
         args = self.prepare_args(*raw_args)
         kwargs = self.prepare_kwargs(**raw_kwargs)
 
+        # Build the cache key and attempt to fetch the cached item
         key = self.key(*args, **kwargs)
         item = cache.get(key)
 
@@ -63,7 +72,7 @@ class Job(object):
             # a) fetch the data immediately, blocking execution until
             #    the fetch has finished, or
             # b) trigger an async refresh and return an empty result
-            if self.should_item_be_fetched_synchronously(*args, **kwargs):
+            if self.should_missing_item_be_fetched_synchronously(*args, **kwargs):
                 logger.debug(("Job %s with key '%s' - cache MISS - running "
                               "synchronous refresh"),
                              self.class_path, key)
@@ -82,22 +91,33 @@ class Job(object):
                 return empty
 
         expiry, data = item
-        if expiry < time.time():
-            # Cache HIT but STALE expiry - we trigger a refresh but allow the
-            # stale result to be returned this time.  This is normally
-            # acceptable.
-            logger.debug(
-                ("Job %s with key '%s' - STALE cache hit - triggering "
-                 "async refresh and returning stale result"),
-                self.class_path, key)
-            # We replace the item in the cache with a 'timeout' expiry - this
-            # prevents cache hammering but guards against a 'limbo' situation
-            # where the refresh task fails for some reason.
-            self.cache_set(key, self.timeout(*args, **kwargs), data)
-            self.async_refresh(*args, **kwargs)
+        delta = time.time() - expiry
+        if delta > 0:
+            # Cache HIT but STALE expiry - we can either:
+            # a) fetch the data immediately, blocking execution until
+            #    the fetch has finished, or
+            # b) trigger a refresh but allow the stale result to be
+            #    returned this time.  This is normally acceptable.
+            if self.should_stale_item_be_fetched_synchronously(
+                    delta, *args, **kwargs):
+                logger.debug(
+                    ("Job %s with key '%s' - STALE cache hit - running "
+                    "synchronous refresh"),
+                    self.class_path, key)
+                return self.refresh(*args, **kwargs)
+            else:
+                logger.debug(
+                    ("Job %s with key '%s' - STALE cache hit - triggering "
+                    "async refresh and returning stale result"),
+                    self.class_path, key)
+                # We replace the item in the cache with a 'timeout' expiry - this
+                # prevents cache hammering but guards against a 'limbo' situation
+                # where the refresh task fails for some reason.
+                timeout = self.timeout(*args, **kwargs)
+                self.cache_set(key, timeout, data)
+                self.async_refresh(*args, **kwargs)
         else:
-            logger.debug(("Job %s with key '%s' - cache HIT"), self.class_path,
-                         key)
+            logger.debug("Job %s with key '%s' - cache HIT", self.class_path, key)
         return data
 
     def invalidate(self, *raw_args, **raw_kwargs):
@@ -114,6 +134,17 @@ class Job(object):
             self.cache_set(key, self.timeout(*args, **kwargs), data)
             self.async_refresh(*args, **kwargs)
 
+    def delete(self, *raw_args, **raw_kwargs):
+        """
+        Remove an item from the cache
+        """
+        args = self.prepare_args(*raw_args)
+        kwargs = self.prepare_kwargs(**raw_kwargs)
+        key = self.key(*args, **kwargs)
+        item = cache.get(key)
+        if item is not None:
+            cache.delete(key)
+
     # --------------
     # HELPER METHODS
     # --------------
@@ -134,14 +165,15 @@ class Job(object):
         """
         cache.set(key, (expiry, data), self.cache_ttl)
 
-        # Warning - not all values save correctly to Memcache, some values
-        # will fail silently.  It's tricky to test for this behaviour as cached
-        # QuerySets aren't "equal" to the original.
-        __, cached_data = cache.get(key, (None, None))
-        if data is not None and cached_data is None:
-            raise RuntimeError(
-                "Unable to save data of type %s to cache" % (
-                    type(data)))
+        if getattr(settings, 'CACHEBACK_VERIFY_CACHE_WRITE', True):
+            # We verify that the item was cached correctly.  This is to avoid a
+            # Memcache problem where some values aren't cached correctly
+            # without warning.
+            __, cached_data = cache.get(key, (None, None))
+            if data is not None and cached_data is None:
+                raise RuntimeError(
+                    "Unable to save data of type %s to cache" % (
+                        type(data)))
 
     def refresh(self, *args, **kwargs):
         """
@@ -169,11 +201,19 @@ class Job(object):
                 call_kwargs=kwargs)
         except Exception, e:
             # Handle exceptions from talking to RabbitMQ - eg connection
-            # refused.
+            # refused.  When this happens, we try to run the task
+            # synchronously.
             logger.error("Unable to trigger task asynchronously - failing "
                          "over to synchronous refresh")
             logger.exception(e)
-            return self.refresh(*args, **kwargs)
+            try:
+                return self.refresh(*args, **kwargs)
+            except Exception, e:
+                # Something went wrong while running the task
+                logger.error("Unable to refresh data synchronously: %s", e)
+                logger.exception(e)
+            else:
+                logger.debug("Failover synchronous refresh completed successfully")
 
     def get_constructor_args(self):
         return ()
@@ -210,12 +250,32 @@ class Job(object):
         """
         return time.time() + self.refresh_timeout
 
-    def should_item_be_fetched_synchronously(self, *args, **kwargs):
+    def should_missing_item_be_fetched_synchronously(self, *args, **kwargs):
         """
-        Return whether to refresh an item synchronously
+        Return whether to refresh an item synchronously when it is missing from
+        the cache
         """
         return self.fetch_on_miss
 
+    def should_item_be_fetched_synchronously(self, *args, **kwargs):
+        import warnings
+        warnings.warn(
+            "The method 'should_item_be_fetched_synchronously' is deprecated "
+            "and will be removed in 0.5.  Use "
+            "'should_missing_item_be_fetched_synchronously' instead.",
+            DeprecationWarning)
+        return self.should_missing_item_be_fetched_synchronously(
+            *args, **kwargs)
+
+    def should_stale_item_be_fetched_synchronously(self, delta, *args, **kwargs):
+        """
+        Return whether to refresh an item synchronously when it is found in the
+        cache but stale
+        """
+        if self.fetch_on_stale_threshold is None:
+            return False
+        return delta > (self.fetch_on_stale_threshold - self.lifetime)
+
     def key(self, *args, **kwargs):
         """
         Return the cache key to use.
